%{
/*  gentrex - generic tree expressions; grammar
    Copyright (C) 2015  Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Project page: http://repo.hu/projects/gentrex
    Author: gentrex@igor2.repo.hu
*/

#include <stdlib.h>
#include <string.h>
#include "gentrex.h"
#include "execute.h"
#include "lex_helper.h"

#define YYLEX_PARAM gtx->scanner, gtx

/* maximum length of a custom operator; malloc() could round this up to 32 perhaps */
#define CUSTOM_OP_MAXLEN 20

void gtx_error();
int gtx_wrap();
int gtx_lex();

%}

%pure-parser
%parse-param {gentrex_state_t *gtx}
%lex-param {void *scanner, gentrex_state_t *gtx}


%union {
	char *str;
	gtx_range_t rng;
	int i;
	gtx_inst_t *ins;
}

%token OP_AND OP_OR OP_EQ INTEGER STRING MATCH_STRING


/*%left ','*/
/*%right '?' ':'*/
%left '!'
%left OP_EQ
%left OP_OR
%left OP_AND
%left '^' '>' '.' '/' ':'
%left MATCH_STRING
%right '='

%type <ins> expr
%type <ins> tag_expr
%type <rng> loop

%type <str> STRING MATCH_STRING
%type <i> INTEGER

%type <ins> OP_AND
%type <ins> OP_OR

%type <str> c_ops
%type <i> c_op
%%

expr:
	/* empty */           {}
	| '['                 { gtx_append(gtx, INST_ENTER); }
	 tag_expr ']'         { $$ = $3; gtx_append(gtx, INST_LEAVE);  }
	| '[' ']'             { $$ = gtx_append(gtx, INST_TAG_TRUE); }
	| STRING              { $$ = gtx_append(gtx, INST_NODE_REF); $$->arg2.str = $1;}
	| '!' expr            { $$ = $2; gtx_append(gtx, INST_NOT); }
	| expr ':' expr       { $$ = $3; } /* NOP: $3 operates on what's left on the stack by $2 */
	| expr '.'            { gtx_append(gtx, INST_TREE_CHILD)->arg1.range.state_idx = -1; }
             expr       { $$ = $4; }
	| expr '.' loop       { gtx_inst_t *i = gtx_append(gtx, INST_TREE_CHILD); memcpy(&i->arg1.range, &$3, sizeof(gtx_range_t)); }
                  expr  { $$ = $5; }
	| expr '/'            { gtx_inst_t *i = gtx_append(gtx, INST_TREE_CHILDREN); i->arg1.range.from = 1;  i->arg1.range.to = 1; i->arg1.range.state_idx = gtx->search_state_len++; }
             expr       { $$ = $4; }
	| expr '/' loop       { gtx_inst_t *i = gtx_append(gtx, INST_TREE_CHILDREN); memcpy(&i->arg1.range, &$3, sizeof(gtx_range_t)); }
	                expr  { $$ = $5; }
	| expr '^'            { gtx_append(gtx, INST_TREE_PARENT)->arg1.range.state_idx = -1; }
             expr       { $$ = $4; }
	| expr '^' loop       { gtx_inst_t *i = gtx_append(gtx, INST_TREE_PARENT); memcpy(&i->arg1.range, &$3, sizeof(gtx_range_t)); }
	                expr  { $$ = $5; }
	| expr '>'            { gtx_append(gtx, INST_TREE_NEXT)->arg1.range.state_idx = -1; }
             expr       { $$ = $4; }
	| expr '>' loop       { gtx_inst_t *i = gtx_append(gtx, INST_TREE_NEXT); memcpy(&i->arg1.range, &$3, sizeof(gtx_range_t)); }
	                expr  { $$ = $5; }
	| expr OP_AND         { $2 = gtx_append(gtx, INST_BRPUSH); $2->arg1.brtype = INST_AND; }
	                expr  { $2->arg2.branch = $4+1; $$=gtx_append(gtx, INST_AND); }
	| expr OP_OR          { $2 = gtx_append(gtx, INST_BRPUSH); $2->arg1.brtype = INST_OR; }
	                expr  { $2->arg2.branch = $4+1; $$=gtx_append(gtx, INST_OR); }
	| STRING '=' expr     { $$ = gtx_append(gtx, INST_ASG); $$->arg2.str = $1; }
	| '('                 { gtx_append(gtx, INST_ENTER); }
	      expr ')'        { $$ = gtx_append(gtx, INST_LEAVE); }
	;

loop:
	  '*'                            { $$.from = 0;  $$.to = GTX_RANGE_INFINITE; $$.state_idx = gtx->search_state_len++; }
	| '+'                            { $$.from = 1;  $$.to = GTX_RANGE_INFINITE; $$.state_idx = gtx->search_state_len++; }
	| '{' INTEGER '}'                { $$.from = $2; $$.to = $2;                 $$.state_idx = gtx->search_state_len++; }
	| '{' INTEGER '-' INTEGER '}'    { $$.from = $2; $$.to = $4;                 $$.state_idx = gtx->search_state_len++; }
	;

tag_expr:
	  '!' tag_expr                   { $$ = gtx_append(gtx, INST_NOT); }
	| STRING OP_EQ STRING            { $$ = gtx_append(gtx, INST_TAG_EQ);    $$->arg1.tagn=$1; $$->arg2.str = $3; $$->match_cookie = gtx->api->op_cookie_alloc == NULL ? NULL : gtx->api->op_cookie_alloc(gtx->api->user_tree_ctx, GTX_OP_EQ,    $$->arg1.tagn, $$->arg2.str, NULL); }
	| STRING MATCH_STRING            { if ($2 == NULL) { yyerror(gtx, "unterminated match argument"); return -1; }
	                                   $$ = gtx_append(gtx, INST_TAG_MATCH); $$->arg1.tagn=$1; $$->arg2.str = $2; $$->match_cookie = gtx->api->op_cookie_alloc == NULL ? NULL : gtx->api->op_cookie_alloc(gtx->api->user_tree_ctx, GTX_OP_MATCH, $$->arg1.tagn, $$->arg2.str, gtx->match_c_op);  $$->c_op = gtx->match_c_op; gtx->match_c_op = NULL; }
	| STRING c_ops STRING            { $$ = gtx_append(gtx, INST_TAG_C_OP);  $$->arg1.tagn=$1; $$->arg2.str = $3; $$->match_cookie = gtx->api->op_cookie_alloc == NULL ? NULL : gtx->api->op_cookie_alloc(gtx->api->user_tree_ctx, GTX_OP_C_OP,  $$->arg1.tagn, $$->arg2.str, $2); $$->c_op = $2; }
	| STRING c_ops INTEGER           { $$ = gtx_append(gtx, INST_TAG_C_OP);  $$->arg1.tagn=$1; $$->arg2.str = strclone(gtx->sval); $$->match_cookie = gtx->api->op_cookie_alloc == NULL ? NULL : gtx->api->op_cookie_alloc(gtx->api->user_tree_ctx, GTX_OP_C_OP,  $$->arg1.tagn, $$->arg2.str, $2); $$->c_op = $2; }
	| tag_expr OP_AND                { $2 = gtx_append(gtx, INST_BRPUSH); $2->arg1.brtype = INST_AND; }
	                 tag_expr        { $2->arg2.branch = $4+1; $$ = gtx_append(gtx, INST_AND); }
	| tag_expr OP_OR                 { $2 = gtx_append(gtx, INST_BRPUSH); $2->arg1.brtype = INST_OR; }
	                 tag_expr        { $2->arg2.branch = $4+1; $$ = gtx_append(gtx, INST_OR); }
	| '(' tag_expr ')'               { $$ = $2; }
	;

c_op:
	  '<'       { $$ = '<'; }
	| '>'       { $$ = '>'; }
	| '='       { $$ = '='; }
	| '!'       { $$ = '!'; }
	| '*'       { $$ = '*'; }
	| '/'       { $$ = '/'; }
	| '^'       { $$ = '^'; }
	| '.'       { $$ = '.'; }
	;

c_ops:
	  c_op           { $$ = malloc(CUSTOM_OP_MAXLEN); $$[0] = $1; $$[1] = '\0'; }
	| c_ops c_op     { int l = strlen($1); $$ = $1; if (l > CUSTOM_OP_MAXLEN-2) { yyerror(gtx, "custom operator too long"); return -1; } $$[l] = $2; $$[l+1] = '\0'; }
	;
